;;;
;;; Step 1 -- Write the entries in each ChangeLog.
;;;

(defun add-titled-change-log-entry (title)
  "Add a ChangeLog entry to the current buffer, with title."
  (interactive "sTitle: ")
  (add-change-log-entry nil buffer-file-name)
  (save-excursion
    (goto-char (point-min))
    (forward-line 1)
    (insert "\n\t" title "\n"))
  ;; Mnemonic for ‘h’ is "head".
  (copy-to-register ?h (point-min) (point)))

(eval-after-load "add-log"
  '(define-key change-log-mode-map
     "\C-c\C-a" 'add-titled-change-log-entry))

;;;
;;; Step 2 -- Merge them into the VCS commit log buffer.
;;;

(require 'log-edit)

(defun compute-commit-log-text (relative-file-names)
  (let ((changelogs (or (remove-if-not (lambda (s)
                                         (string-match "ChangeLog$" s))
                                       relative-file-names)
                        (error "No ChangeLog files")))
        head body)
    (dolist (fn changelogs)
      (with-temp-buffer
        (insert-file-contents fn)
        (search-forward "\n\n" nil nil 2)
        (push (buffer-substring (point-min) (point)) head)
        (unless (or (null (cdr head))
                    (string= (car head) (cadr head)))
          (error "head mismatch for %s" fn))
        (push (buffer-substring
               (point) (progn (re-search-forward change-log-start-entry-re)
                              (match-beginning 0)))
              body)))
    (setq head (car head)
          head (and (string-match "\n\n\t" head)
                    (substring head (match-end 0)))
          body (map 'list
                    (lambda (fn bod)
                      (let ((pre (and (string-match "ChangeLog$" fn)
                                      (substring fn 0 (match-beginning 0)))))
                        (with-temp-buffer
                          (insert bod)
                          (goto-char (point-min))
                          (while (search-forward "\t* " nil 1)
                            (insert pre))
                          (while (re-search-backward "^\t" nil t)
                            (delete-char 1))
                          (buffer-string))))
                    (nreverse changelogs) body))
    (cons head body)))

;; Not actually Git-specific, although untested with any other VCS.
;; The "scan" is for an emacsclient-based workflow from 2008, now defunct.
(defun git-scan-log-buffer-and-insert-commit-log ()
  (interactive)
  (apply 'insert (compute-commit-log-text log-edit-initial-files)))

(define-key log-edit-mode-map "\C-c\C-d"
  'git-scan-log-buffer-and-insert-commit-log)

;;;
;;; Step 3 -- Resynchronize Emacs.
;;;

;;;###autoload
(defun revert-some-buffers (&optional skip-read-only)
  "Revert each file-related buffer.
Preifx arg means don't bother with read-only buffers."
  (interactive "P")
  (save-some-buffers)
  (let ((ready (remove-if-not (lambda (buf)
                                (with-current-buffer buf
                                  (and buffer-file-name
                                       (or (not skip-read-only)
                                           (not buffer-read-only)))))
                              (buffer-list))))
    (save-excursion
      (dolist (buf ready)
        (set-buffer buf)
        (revert-buffer t t)))
    (message "Buffers reverted: %d %s" (length ready) ready)))

(global-set-key [(control c) (control meta r)] 'revert-some-buffers)
